import os

from bravado.exception import HTTPError
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required, user_passes_test
from django.core.exceptions import PermissionDenied
from django.db import IntegrityError
from django.db.models import Count
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.translation import ugettext_lazy as _
from esi.decorators import token_required
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.http import HttpResponse
import csv
import re
from itertools import chain

from .models import CorpStat, CorpMember


def access_corpstats_test(user):
    return user.has_perm('corpstats.view_corp_corpstats') \
           or user.has_perm('corpstats.view_alliance_corpstats') \
           or user.has_perm('corpstats.view_state_corpstats') \
           or user.has_perm('corpstats.add_corpstats')


def corpstats_visible_to_user(view):
    def check_corpstats(request, corp_id=None):
        if corp_id:
            corp = get_object_or_404(EveCorporationInfo, corporation_id=corp_id)
            corpstats = get_object_or_404(CorpStat, corp=corp)

            # get available models
            available = CorpStat.objects.visible_to(request.user)

            # ensure we can see the requested model
            if corpstats and corpstats not in available:
                raise PermissionDenied('You do not have permission to view the selected corporation statistics module.')
        else:
            corpstats = None
        return view(request, corpstats, corp_id=corp_id)
    return check_corpstats


@login_required
@user_passes_test(access_corpstats_test)
@permission_required('corpstats.add_corpstats')
@token_required(scopes=['esi-corporations.track_members.v1', 'esi-universe.read_structures.v1'])
def corpstats_add(request, token):
    try:
        if EveCharacter.objects.filter(character_id=token.character_id).exists():
            corp_id = EveCharacter.objects.get(character_id=token.character_id).corporation_id
        else:
            corp_id = \
                token.get_esi_client().Character.get_characters_character_id(
                    character_id=token.character_id).result()['corporation_id']
        try:
            corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
        except EveCorporationInfo.DoesNotExist:
            corp = EveCorporationInfo.objects.create_corporation(corp_id)
        cs = CorpStat.objects.create(token=token, corp=corp)
        try:
            cs.update()
        except HTTPError as e:
            messages.error(request, str(e))
        assert cs.pk  # ensure update was successful
        if CorpStat.objects.filter(pk=cs.pk).visible_to(request.user).exists():
            return redirect('corpstat:view_corp', corp_id=corp.corporation_id)
    except IntegrityError:
        messages.error(request, _('Selected corp already has a statistics module.'))
    except AssertionError:
        messages.error(request, _('Failed to gather corporation statistics with selected token.'))
    return redirect('corpstat:view')


@login_required
@user_passes_test(access_corpstats_test)
@corpstats_visible_to_user
def corpstat_view(request, corpstats, corp_id=None,):
    # get available models
    available = CorpStat.objects.visible_to(request.user).order_by('corp__corporation_name')

    # get default model if none requested
    if not corp_id and available.count() == 1:
        corpstats = available[0]
    elif not corp_id and available.count() > 1 and request.user.profile.main_character:
        # get their main corp if available
        try:
            corpstats = available.get(corp__corporation_id=request.user.profile.main_character.corporation_id)
        except CorpStats.DoesNotExist:
            pass

    context = {
        'available_corps': available,
        'available_alliances': CorpStat.objects.alliances_visible_to(request.user),
    }
    if corpstats:
        mains, members, unregistered, tracking = corpstats.get_stats()
        context.update({
            'mains': mains,
            'members': members,
            'unregistered_members': unregistered,
            'tracking': tracking,
            'corpstats': corpstats,
        })

    return render(request, 'corpstat/corpstats.html', context=context)


@login_required
@user_passes_test(access_corpstats_test)
@corpstats_visible_to_user
def corpstats_update(request, corpstats, **_):
    try:
        corpstats.update()
    except HTTPError as e:
        messages.error(request, str(e))
    if corpstats.pk:
        return redirect('corpstat:view_corp', corp_id=corpstats.corp.corporation_id)
    else:
        return redirect('corpstat:view')


@login_required
@user_passes_test(access_corpstats_test)
def corpstats_search(request):
    results = []
    search_string = request.GET.get('search_string', None)
    if search_string:
        has_similar = CorpStat.objects.filter(members__character_name__icontains=search_string).visible_to(
            request.user).distinct()
        for corpstats in has_similar:
            similar = corpstats.members.filter(character_name__icontains=search_string)
            for s in similar:
                results.append((corpstats, s))
        results = sorted(results, key=lambda x: x[1].character_name)
        context = {
            'available_corps': CorpStat.objects.visible_to(request.user),
            'available_alliances': CorpStat.objects.alliances_visible_to(request.user),
            'results': results,
            'search_string': search_string,
        }
        return render(request, 'corpstat/search.html', context=context)
    return redirect('corpstat:view')


@login_required
@user_passes_test(access_corpstats_test)
def alliance_view(request, alliance_id=None):
    # get available models
    alliances = CorpStat.objects.alliances_visible_to(request.user)

    # get default model if none requested
    if not alliance_id:
        alliance_id = request.user.profile.main_character.alliance.alliance_id
    """
    # ensure we can see the requested model
    if alliance_id not in alliances:
        raise PermissionDenied('You do not have permission to view the selected alliance statistics module.')
    """
    corpstats_temp = CorpMember.objects.values('corpstats__corp__corporation_name').annotate(total_members=Count('character_id')).order_by('corpstats__corp__corporation_name') 
    corp_totals={}
    for stat in corpstats_temp:
        corp_totals[stat['corpstats__corp__corporation_name']]=stat['total_members']
    # it's a lot easier to count member objects directly than try to walk reverse relations
    alliance_members = EveCharacter.objects.filter(character_ownership__user__profile__main_character__alliance_id=alliance_id).filter(alliance_id=alliance_id)
    corp_breakdown = {}
    
    for member in alliance_members:
        if member.corporation_name not in corp_breakdown:
            corp_breakdown[member.corporation_name] = {}
            corp_breakdown[member.corporation_name]['mains'] = 0
            corp_breakdown[member.corporation_name]['alts'] = 0
            corp_breakdown[member.corporation_name]['members'] = 0
            try: 
                corp_breakdown[member.corporation_name]['total'] = corp_totals[member.corporation_name]
            except:
                pass

        if member.character_ownership.user.profile.main_character == member:
            corp_breakdown[member.corporation_name]['mains'] += 1
        else:
            corp_breakdown[member.corporation_name]['alts'] += 1 
        corp_breakdown[member.corporation_name]['members'] += 1


    context = {
        'available_corps': CorpStat.objects.visible_to(request.user),
        'available_alliances': alliances,
        'alliance_id': alliance_id,
        'alliance_name': alliances[alliance_id],
        'corp_breakdown': corp_breakdown
    }

    return render(request, 'corpstat/alliancestats.html', context=context)


@login_required
@user_passes_test(access_corpstats_test)
@corpstats_visible_to_user
def export_corpstats(request, corpstats, **_):
    if not corpstats.members.all().exists():
        # there are no members, say there's no data
        return HttpResponse(status=204)

    many_to_many_field_names = set([many_to_many_field.name for many_to_many_field in CorpMember._meta.many_to_many])
    field_names = [field.name for field in CorpMember._meta.get_fields() if not field.auto_created
                   and field.name != 'corpstats' and field.name not in many_to_many_field_names]

    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = 'attachment; filename=%s.csv' % re.sub('[^\w]', '', corpstats.corp.corporation_name)

    writer = csv.writer(response)
    writer.writerow(list(chain(field_names, many_to_many_field_names)))

    for member in corpstats.members.all().order_by('character_name'):
        row = []
        for field in field_names:
            row.append('' if getattr(member, field) is None else str(getattr(member, field)))
        for field in many_to_many_field_names:
            row.extend([str(a) for a in getattr(member, field).all()])
        writer.writerow(row)

    return response



import os
from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')

from django.conf import settings  # noqa

app = Celery('devauth')

# Using a string here means the worker don't have to serialize
# the configuration object to child processes.
app.config_from_object('django.conf:settings')
app.conf.ONCE = {
    'backend': 'allianceauth.services.tasks.DjangoBackend',
    'settings': {}
}

# Load task modules from all registered Django app configs.
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)




# Generated by Django 2.2 on 2019-11-12 14:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('esi', '0005_remove_token_length_limit'),
        ('eveonline', '0010_alliance_ticker'),
    ]

    operations = [
        migrations.CreateModel(
            name='CorpStat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('corp', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='eveonline.EveCorporationInfo')),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='esi.Token')),
            ],
            options={
                'verbose_name': 'corp stats',
                'verbose_name_plural': 'corp stats',
                'permissions': (('view_corp_corpstats', 'Can view corp stats of their corporation.'), ('view_alliance_corpstats', 'Can view corp stats of members of their alliance.'), ('view_state_corpstats', 'Can view corp stats of members of their auth state.'), ('view_all_corpstats', 'Can view all corp stats.')),
            },
        ),
        migrations.CreateModel(
            name='CorpMember',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('character_id', models.PositiveIntegerField()),
                ('character_name', models.CharField(max_length=50)),
                ('location_id', models.BigIntegerField()),
                ('location_name', models.CharField(blank=True, max_length=150, null=True)),
                ('ship_type_id', models.PositiveIntegerField()),
                ('ship_type_name', models.CharField(max_length=42)),
                ('start_date', models.DateTimeField()),
                ('logon_date', models.DateTimeField()),
                ('logoff_date', models.DateTimeField()),
                ('base_id', models.PositiveIntegerField(blank=True, null=True)),
                ('corpstats', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='members', to='corpstats.CorpStat')),
            ],
            options={
                'ordering': ['character_name'],
                'unique_together': {('corpstats', 'character_id')},
            },
        ),
    ]



from django.db import models
import logging

logger = logging.getLogger(__name__)


class CorpStatQuerySet(models.QuerySet):
    def visible_to(self, user):

        if user.has_perm('corpstats.view_all_corpstats'):  # superusers and users with this permission
            logger.debug('Returning all corpstats for %s.' % user)
            return self

        try:
            char = user.profile.main_character
            assert char
            # build all accepted queries
            queries = [models.Q(token__user=user)]

            if user.has_perm('corpstats.view_corp_corpstats'):
                queries.append(models.Q(corp__corporation_id=char.corporation_id))
            if user.has_perm('corpstats.view_alliance_corpstats'):
                queries.append(models.Q(corp__alliance_id=char.alliance_id))
            if user.has_perm('corpstats.view_state_corpstats'):
                queries.append(models.Q(corp__in=user.profile.state.member_corporations.all()))
                queries.append(models.Q(
                    corp__alliance_id__in=user.profile.state.member_alliances.all().values_list('alliance_id',
                                                                                                flat=True)))
            logger.debug('%s queries for user %s visible corpstats.' % (len(queries), user))
            # filter based on queries
            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug('User %s has no main character. No corpstats visible.' % user)
            return self.none()


class CorpStatManager(models.Manager):
    def get_queryset(self):
        return CorpStatQuerySet(self.model, using=self._db)

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def alliances_visible_to(self, user):
        alliances = []
        if user.has_perm('corpstats.view_all_corpstats'):
            alliances = set(self.visible_to(user).filter(corp__alliance__alliance_id__isnull=False).values_list('corp__alliance__alliance_id', 'corp__alliance__alliance_name'))
        elif user.profile.main_character:
            if user.has_perm('corpstats.view_alliance_corpstats') and user.profile.main_character.alliance_id:
                alliances.append((user.profile.main_character.alliance_id, user.profile.main_character.alliance_name))
            if user.has_perm('corpstats.view_state_corpstats'):
                alliances += [(a.alliance_id, a.alliance_name) for a in user.profile.state.member_alliances.all()]
        return {a[0]: a[1] for a in alliances}



from unittest import mock

from django.test import TestCase
from django.utils.timezone import now
from allianceauth.tests.auth_utils import AuthUtils
from .models import CorpStat, CorpMember
from allianceauth.eveonline.models import EveCorporationInfo, EveAllianceInfo, EveCharacter
from esi.models import Token
from esi.errors import TokenError
from bravado.exception import HTTPForbidden
from django.contrib.auth.models import User, Permission
from allianceauth.authentication.models import CharacterOwnership

class CorpStatsManagerTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST')
        cls.user.profile.refresh_from_db()
        cls.user2 = AuthUtils.create_user('test2')
        AuthUtils.add_main_character(cls.user2, 'another test character', '5', corp_id='4', corp_name='another_test_corp', corp_ticker='TEST2', alliance_id='6', alliance_name='TEST2')
        cls.user2.profile.refresh_from_db()
        cls.user3 = AuthUtils.create_user('test3')
        AuthUtils.add_main_character(cls.user2, 'yet_another test character', '9', corp_id='7', corp_name='yet_another_test_corp', corp_ticker='TEST3')

        cls.user2.profile.refresh_from_db()
        cls.alliance = EveAllianceInfo.objects.create(alliance_id='3', alliance_name='test alliance', alliance_ticker='TEST', executor_corp_id='2')
        cls.corp = EveCorporationInfo.objects.create(corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', alliance=cls.alliance, member_count=1)
        cls.alliance2 = EveAllianceInfo.objects.create(alliance_id='6', alliance_name='another test alliance', alliance_ticker='TEST2', executor_corp_id='4')
        cls.corp2 = EveCorporationInfo.objects.create(corporation_id='4', corporation_name='another test corp', corporation_ticker='TEST2', alliance=cls.alliance2, member_count=1)
        cls.corp3 = EveCorporationInfo.objects.create(corporation_id='7', corporation_name='yet_another test corp', corporation_ticker='TEST3', alliance=None, member_count=1)
        cls.token = Token.objects.create(user=cls.user, access_token='a', character_id='1', character_name='test character', character_owner_hash='z')
        cls.corpstat = CorpStat.objects.create(corp=cls.corp, token=cls.token)
        cls.token2 = Token.objects.create(user=cls.user2, access_token='b', character_id='5', character_name='another test character', character_owner_hash='y')
        cls.corpstat2 = CorpStat.objects.create(corp=cls.corp2, token=cls.token2)
        cls.token3 = Token.objects.create(user=cls.user3, access_token='c', character_id='9', character_name='yet_another test character', character_owner_hash='x')
        cls.corpstat3 = CorpStat.objects.create(corp=cls.corp3, token=cls.token3)
        cls.view_all_corp_permission = Permission.objects.get_by_natural_key('view_all_corpstats', 'corpstats', 'corpstat')
        cls.view_corp_permission = Permission.objects.get_by_natural_key('view_corp_corpstats', 'corpstats', 'corpstat')
        cls.view_alliance_permission = Permission.objects.get_by_natural_key('view_alliance_corpstats', 'corpstats', 'corpstat')
        cls.view_state_permission = Permission.objects.get_by_natural_key('view_state_corpstats', 'corpstats', 'corpstat')
        cls.state = AuthUtils.create_state('test state', 500)

        AuthUtils.assign_state(cls.user, cls.state, disconnect_signals=True)

    def setUp(self):
        self.user.refresh_from_db()
        self.user.user_permissions.clear()
        self.user2.refresh_from_db()
        self.user2.user_permissions.clear()
        self.state.refresh_from_db()
        self.state.member_corporations.clear()
        self.state.member_alliances.clear()
        self.user.is_superuser = False

    def test_visible_corporation(self):
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_corp_permission)
        cs = CorpStat.objects.visible_to(user)
        self.assertIn(self.corpstat, cs)
        self.assertNotIn(self.corpstat2, cs)

    def test_visible_alliance(self):
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_alliance_permission)
        cs = CorpStat.objects.visible_to(user)
        self.assertIn(self.corpstat, cs)
        self.assertNotIn(self.corpstat2, cs)

    def test_visible_state_corp_member(self):
        self.state.member_corporations.add(self.corp)
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_state_permission)
        cs = CorpStat.objects.visible_to(user)
        self.assertIn(self.corpstat, cs)
        self.assertNotIn(self.corpstat2, cs)

    def test_visible_state_alliance_member(self):
        self.state.member_alliances.add(self.alliance)
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_state_permission)
        cs = CorpStat.objects.visible_to(user)
        self.assertIn(self.corpstat, cs)
        self.assertNotIn(self.corpstat2, cs)

    def test_corp_visible_alliances(self):
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_corp_permission)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertEquals(len(alliances), 0)

    def test_alliance_visible_alliances(self):
        user = User.objects.get(pk=self.user.pk)  # permissions cache is only cleared when retrieved fresh from db
        user.user_permissions.add(self.view_alliance_permission)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertIn('3', alliances)
        self.assertNotIn('6', alliances)

    def test_no_perms_visible_alliances(self):
        user = User.objects.get(pk=self.user.pk)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertEquals(len(alliances), 0)

    def test_empty_state_visible_alliances(self):
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_state_permission)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertEquals(len(alliances), 0)

    def test_view_all_visible_alliances(self):
        user = User.objects.get(pk=self.user.pk)
        user.user_permissions.add(self.view_all_corp_permission)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertIn('3', alliances)
        self.assertIn('6', alliances)
        self.assertEquals(len(alliances), 2)

    def test_state_visible_to(self):
        user = User.objects.get(pk=self.user.pk)
        self.state.member_alliances.add(self.alliance)
        user.user_permissions.add(self.view_state_permission)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertIn('3', alliances)
        self.assertNotIn('6', alliances)

    def test_visible_superuser(self):
        user = User.objects.get(pk=self.user.pk)
        user.is_superuser = True
        cs = CorpStat.objects.visible_to(user)
        self.assertIn(self.corpstat, cs)
        self.assertIn(self.corpstat2, cs)

    def test_no_main_visible_superuser(self):
        user = User.objects.get(pk=self.user3.pk)
        alliances = CorpStat.objects.alliances_visible_to(user)
        self.assertEquals(len(alliances), 0)

class CorpStatsUpdateTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST')
        cls.token = Token.objects.create(user=cls.user, access_token='a', character_id='1', character_name='test character', character_owner_hash='z')
        cls.corp = EveCorporationInfo.objects.create(corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', member_count=1)

    def setUp(self):
        self.corpstat = CorpStat.objects.get_or_create(token=self.token, corp=self.corp)[0]

    def test_can_update(self):
        self.assertTrue(self.corpstat.can_update(self.user))
        self.corpstat.token.user = None
        self.assertFalse(self.corpstat.can_update(self.user))
        self.user.is_superuser = True
        self.assertTrue(self.corpstat.can_update(self.user))
        self.user.refresh_from_db()
        self.corpstat.token.refresh_from_db()

    @mock.patch('esi.clients.SwaggerClient')
    def test_update_add_member(self, SwaggerClient):
        SwaggerClient.return_value.Character.get_characters_character_id.return_value.result.return_value = {'corporation_id': 2}
        SwaggerClient.return_value.Corporation.get_corporations_corporation_id_membertracking.return_value.result.return_value = [
            {'character_id': 1, 'ship_type_id': 2, 'location_id': 3, 'logon_date': now(), 'logoff_date': now(), 'start_date': now()}]
        SwaggerClient.return_value.Universe.get_universe_types_type_id.return_value.result.return_value = {'name': 'test ship'}
        SwaggerClient.return_value.Universe.post_universe_names.return_value.result.return_value = [{'id': 1, 'name': 'test character', 'category':'character'}]

        self.corpstat.update()
        self.assertTrue(CorpMember.objects.filter(character_id=1, character_name='test character', corpstats=self.corpstat).exists())

    @mock.patch('esi.clients.SwaggerClient')
    def test_update_remove_member(self, SwaggerClient):
        CorpMember.objects.create(character_id='2', character_name='old test character', corpstats=self.corpstat, location_id=1, location_name='test', ship_type_id=1, ship_type_name='test', logoff_date=now(), logon_date=now(), start_date=now())
        SwaggerClient.return_value.Character.get_characters_character_id.return_value.result.return_value = {'corporation_id': 2}
        SwaggerClient.return_value.Corporation.get_corporations_corporation_id_membertracking.return_value.result.return_value = [{'character_id': 1, 'ship_type_id': 2, 'location_id': 3, 'logon_date': now(), 'logoff_date': now(), 'start_date': now()}]
        SwaggerClient.return_value.Universe.get_universe_types_type_id.return_value.result.return_value = {'name': 'test ship'}
        SwaggerClient.return_value.Universe.post_universe_names.return_value.result.return_value = [{'id': 1, 'name': 'test character', 'category':'character'}]
        self.corpstat.update()
        self.assertFalse(CorpMember.objects.filter(character_id='2', corpstats=self.corpstat).exists())

    @mock.patch('corpstats.models.notify')
    @mock.patch('esi.clients.SwaggerClient')
    def test_update_deleted_token(self, SwaggerClient, notify):
        SwaggerClient.return_value.Character.get_characters_character_id.return_value.result.return_value = {'corporation_id': 2}
        SwaggerClient.return_value.Corporation.get_corporations_corporation_id_membertracking.return_value.result.side_effect = TokenError()
        self.corpstat.update()
        self.assertFalse(CorpStat.objects.filter(corp=self.corp).exists())
        self.assertTrue(notify.called)

    @mock.patch('corpstats.models.notify')
    @mock.patch('esi.clients.SwaggerClient')
    def test_update_http_forbidden(self, SwaggerClient, notify):
        SwaggerClient.return_value.Character.get_characters_character_id.return_value.result.return_value = {'corporation_id': 2}
        SwaggerClient.return_value.Corporation.get_corporations_corporation_id_membertracking.return_value.result.side_effect = HTTPForbidden(mock.Mock())
        self.corpstat.update()
        self.assertFalse(CorpStat.objects.filter(corp=self.corp).exists())
        self.assertTrue(notify.called)

    @mock.patch('corpstats.models.notify')
    @mock.patch('esi.clients.SwaggerClient')
    def test_update_token_character_corp_changed(self, SwaggerClient, notify):
        SwaggerClient.return_value.Character.get_characters_character_id.return_value.result.return_value = {'corporation_id': 3}
        self.corpstat.update()
        self.assertFalse(CorpStat.objects.filter(corp=self.corp).exists())
        self.assertTrue(notify.called)


class CorpStatsPropertiesTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_name='TEST')
        cls.user.profile.refresh_from_db()
        cls.token = Token.objects.create(user=cls.user, access_token='a', character_id='1', character_name='test character', character_owner_hash='z')
        cls.alliance = EveAllianceInfo.objects.create(alliance_id=3, alliance_name='test alliance', alliance_ticker='TEST', executor_corp_id='2')
        cls.corp = EveCorporationInfo.objects.create(corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', alliance_id=3, member_count=1)
        cls.corp.alliance = cls.alliance
        cls.corp.save()
        cls.corpstat = CorpStat.objects.create(token=cls.token, corp=cls.corp)
        cls.character = EveCharacter.objects.create(character_name='another test character', character_id='4', corporation_id='2', corporation_name='test corp', corporation_ticker='TEST')
        AuthUtils.disconnect_signals()
        CharacterOwnership.objects.create(character=cls.character, user=cls.user, owner_hash='a')
        AuthUtils.connect_signals()

    def test_logos(self):
        self.assertEqual(self.corpstat.corp_logo(size=128), 'https://image.eveonline.com/Corporation/2_128.png')
        self.assertEqual(self.corpstat.alliance_logo(size=128), 'https://image.eveonline.com/Alliance/3_128.png')


class CorpMemberTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('test')
        AuthUtils.add_main_character(cls.user, 'test character', '1', corp_id='2', corp_name='test_corp', corp_ticker='TEST', alliance_id='3', alliance_name='TEST')
        cls.user.profile.refresh_from_db()
        cls.token = Token.objects.create(user=cls.user, access_token='a', character_id='1', character_name='test character', character_owner_hash='a')
        cls.alliance = EveAllianceInfo.objects.create(alliance_id='3', alliance_name='test alliance', alliance_ticker='TEST', executor_corp_id='2')
        cls.corp = EveCorporationInfo.objects.create(corporation_id='2', corporation_name='test corp', corporation_ticker='TEST', alliance_id=3, member_count=1)
        cls.corpstat = CorpStat.objects.create(token=cls.token, corp=cls.corp)
        cls.member = CorpMember.objects.create(corpstats=cls.corpstat, character_id='2', character_name='other test character', location_id=1, location_name='test', ship_type_id=1, ship_type_name='test', logoff_date=now(), logon_date=now(), start_date=now())

    def test_portrait_url(self):
        self.assertEquals(self.member.portrait_url(size=32), 'https://image.eveonline.com/Character/2_32.jpg')
        self.assertEquals(self.member.portrait_url(size=32), self.member.portrait_url_32)



from django.conf.urls import url

from . import views

app_name = 'corpstat'

urlpatterns = [
    url(r'^$', views.corpstat_view, name='view'),
    url(r'^add/$', views.corpstats_add, name='add'),
    url(r'^alliance/(?P<alliance_id>(\d)+)/$', views.alliance_view, name='view_alliance'),
    url(r'^(?P<corp_id>(\d)*)/$', views.corpstat_view, name='view_corp'),
    url(r'^(?P<corp_id>(\d)+)/update/$', views.corpstats_update, name='update'),
    url(r'^(?P<corp_id>(\d)+)/export/$', views.export_corpstats, name='export'), # has no permissions
    url(r'^search/$', views.corpstats_search, name='search'),
    ]



from celery import shared_task
from .models import CorpStat


@shared_task
def update_corpstats(pk):
    cs = CorpStat.objects.get(pk=pk)
    cs.update()


@shared_task
def update_all_corpstats():
    for cs in CorpStat.objects.all():
        update_corpstats.delay(cs.pk)



from django.conf.urls import url

import allianceauth.urls
from . import views

urlpatterns = allianceauth.urls.urlpatterns

urlpatterns += [
    # Navhelper test urls
    url(r'^main-page/$', views.page, name='p1'),
    url(r'^main-page/sub-section/$', views.page, name='p1-s1'),
    url(r'^second-page/$', views.page, name='p1'),
]




import logging
import os

from allianceauth.authentication.models import CharacterOwnership, UserProfile
from bravado.exception import HTTPForbidden
from django.db import models
from django.core.exceptions import ObjectDoesNotExist

from esi.errors import TokenError
from esi.models import Token
from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter
from allianceauth.notifications import notify

from .managers import CorpStatManager

logger = logging.getLogger(__name__)


class CorpStat(models.Model):
    token = models.ForeignKey(Token, on_delete=models.CASCADE)
    corp = models.OneToOneField(EveCorporationInfo, on_delete=models.CASCADE)
    last_update = models.DateTimeField(auto_now=True)

    class Meta:
        permissions = (
            ('view_corp_corpstats', 'Can view corp stats of their corporation.'),
            ('view_alliance_corpstats', 'Can view corp stats of members of their alliance.'),
            ('view_state_corpstats', 'Can view corp stats of members of their auth state.'),
            ('view_all_corpstats', 'Can view all corp stats.'),
        )
        verbose_name = "corp stats"
        verbose_name_plural = "corp stats"

    objects = CorpStatManager()

    def __str__(self):
        return "%s for %s" % (self.__class__.__name__, self.corp)

    def update(self):
        try:
            c = self.token.get_esi_client()

            # make sure the token owner is still in this corp

            assert c.Character.get_characters_character_id(
                character_id=self.token.character_id).result()['corporation_id'] == int(self.corp.corporation_id)

            # get member tracking data and retrieve member ids for translation
            tracking = c.Corporation.get_corporations_corporation_id_membertracking(corporation_id=self.corp.corporation_id).result()
            member_ids = [t['character_id'] for t in tracking]

            # requesting too many ids per call results in a HTTP400
            # the swagger spec doesn't have a maxItems count
            # manual testing says we can do over 350, but let's not risk it
            member_id_chunks = [member_ids[i:i + 255] for i in range(0, len(member_ids), 255)]
            member_name_chunks = [c.Universe.post_universe_names(ids=id_chunk).result() for id_chunk in
                                  member_id_chunks]

            member_list = {t['character_id']: t for t in tracking}
            for name_chunk in member_name_chunks:
                for name in name_chunk:
                    member_list[name['id']]['character_name'] = name.get('name', "")

            # get ship and location names
            for t in tracking:
                t['ship_type_name'] = c.Universe.get_universe_types_type_id(type_id=t['ship_type_id']).result()['name'] #TODO use the inbuilt eve provider
                #locations = c.Universe.post_universe_names(ids=[t['location_id']]).result()
                #t['location_name'] = locations[0]['name'] if locations else ''  # might be a citadel we can't know about
                member_list[t['character_id']].update(t)

            # purge old members
            CorpMember.objects.filter(corpstats=self).exclude(character_id__in=member_ids).delete()

            # bulk update and create new member models
            for c_id, data in member_list.items():
                CorpMember.objects.update_or_create(character_id=c_id, corpstats=self, defaults=data)

            # update the timer
            self.save()

        except TokenError as e:
            logger.warning("%s failed to update: %s" % (self, e))
            if self.token.user:
                notify(self.token.user, "%s failed to update with your ESI token." % self,
                       message="Your token has expired or is no longer valid. Please add a new one to create a new CorpStats.",
                       level="error")
            self.delete()
        except HTTPForbidden as e:
            logger.warning("%s failed to update: %s" % (self, e))
            if self.token.user:
                notify(self.token.user, "%s failed to update with your ESI token." % self,
                       message="%s: %s" % (e.status_code, e.message), level="error")
            self.delete()
        except AssertionError:
            logger.warning("%s token character no longer in corp." % self)
            if self.token.user:
                notify(self.token.user, "%s cannot update with your ESI token." % self,
                       message="%s cannot update with your ESI token as you have left corp." % self, level="error")
            self.delete()

    def get_stats(self):
        """
        return all corpstats for corp

        :return:
        Mains with Alts Dict
        Members List[EveCharacter]
        Un-registered QuerySet[CorpMember]
        """
        character_list = CorpMember.objects.filter(corpstats=self)
        linked_chars = EveCharacter.objects.filter(character_id__in=character_list.values_list('character_id', flat=True))
        linked_chars = linked_chars | EveCharacter.objects.filter( character_ownership__user__profile__main_character__corporation_id=self.corp.corporation_id)
        linked_chars = linked_chars.select_related('character_ownership','character_ownership__user__profile__main_character') \
            .prefetch_related('character_ownership__user__character_ownerships') \
            .prefetch_related('character_ownership__user__character_ownerships__character')

        members = []
        mains = {}

        temp_ids = []
        for char in linked_chars:
            try:
                main = char.character_ownership.user.profile.main_character # get main
                if main is not None:
                    if main.character_id in mains:
                        mains[main.character_id]['alts'].append(char) # append alt
                    else:
                        mains[main.character_id] = {} # create dict
                        mains[main.character_id]['alts'] = []
                        mains[main.character_id]['main'] = main
                        mains[main.character_id]['alts'].append(char)
                    if char.corporation_id == self.corp.corporation_id:
                        members.append(char) # add to member list
                    temp_ids.append(char.character_id) # ignore this char for un-reg'd
            except ObjectDoesNotExist:
                # character has no link
                pass

        unregistered = character_list.exclude(character_id__in=temp_ids)
        tracking = character_list.filter(character_id__in=temp_ids)
        #print(mains, flush=True)
        #print(members, flush=True)
        #print(unregistered, flush=True)

        return mains, members, unregistered, tracking

    def visible_to(self, user):
        return CorpStat.objects.filter(pk=self.pk).visible_to(user).exists()

    def can_update(self, user):
        return self.token.user == user or self.visible_to(user)

    def corp_logo(self, size=128):
        return "https://image.eveonline.com/Corporation/%s_%s.png" % (self.corp.corporation_id, size)

    def alliance_logo(self, size=128):
        if self.corp.alliance_id:
            return "https://image.eveonline.com/Alliance/%s_%s.png" % (self.corp.alliance.alliance_id, size)
        else:
            return "https://image.eveonline.com/Alliance/1_%s.png" % size


class CorpMember(models.Model):
    character_id = models.PositiveIntegerField()
    character_name = models.CharField(max_length=50)  # allegedly

    location_id = models.BigIntegerField()
    location_name = models.CharField(blank=True, null=True, max_length=150)  # this was counted

    ship_type_id = models.PositiveIntegerField()
    ship_type_name = models.CharField(max_length=42)  # this was also counted

    start_date = models.DateTimeField()
    logon_date = models.DateTimeField()
    logoff_date = models.DateTimeField()

    base_id = models.PositiveIntegerField(blank=True, null=True)

    corpstats = models.ForeignKey(CorpStat, on_delete=models.CASCADE, related_name='members')

    class Meta:
        # not making character_id unique in case a character moves between two corps while only one updates
        unique_together = ('corpstats', 'character_id')
        ordering = ['character_name']

    def __str__(self):
        return self.character_name

    def portrait_url(self, size=32):
        return "https://image.eveonline.com/Character/%s_%s.jpg" % (self.character_id, size)

    def __getattr__(self, item):
        if item.startswith('portrait_url_'):
            size = item.strip('portrait_url_')
            return self.portrait_url(size)
        return self.__getattribute__(item)
